#ifndef __MACROS_H__
#define __MACROS_H__

#define U64(n) (uint64_t) n
#define I64(n) (int64_t) n
#define U32(n) (uint32_t) n
#define I32(n) (int32_t) n
#define U16(n) (uint16_t) n
#define I16(n) (int16_t) n
#define U8(n)  (uint8_t) n
#define I8(n)  (int8_t) n

#define MEMSIZE U64

#endif
